%---------------------------------------------------------------
%
%  Function rdr_convert
%
%  Written by Graeme Martin, graemem@ssec.wisc.edu
%  Copyright (c) 2007-2017 University of Wisconsin Regents.
%
%  This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
%  Parts of this program are based on Dan Mooney's CrIS data
%    analysis tool, (c) Copyright 2004-2013 Massachusetts Institute of
%    Technology
%
%  Description:
%
%    Reads raw CrIS RDR packets from a binary file and unscales the individual fields
%    in each packet as specified in the XML database. 
%
%    Returns:
%      - unpacked interferograms and associated data (FOR, sweep direction, time)
%      - metadata
%      - raw packet data
%
%    Data can optionally can be written to a .mat file.
% 
%    See the README for more information about the functionality and
%    limitations of this release.
% 
%  Calling syntax:
%
%    With only the required arguments specified:
%
%    [ifgs, metadata, packets] = rdr_convert(packets_filename, xml_database_version, side)
%
%    With all optional arguments specified:
%
%    [ifgs, metadata, packets] = rdr_convert(packets_filename, xml_database_version, side, ...
%      write_mat_file, output_directory, bit_trim_in_filename, bit_trim_out_filename, ...
%      packet_format)
%
%  Required arguments:
%
%    packets_filename      The name of the input binary file containing packets
%    xml_database_version  Version of the XML database that describes the
%                          packets in the input file. Currently supported 
%                          values are '1.56', '1.69', 'draft_j1', 'j1_55_005', and 
%                          'j1_55a_008', 'j2_OO_12'. The correct value for operational SNPP 
%                          data should be '1.69'. The correct value for J1 should be 
%                          'j1_55a_008'.
%    side                  Sensor side, either 1 or 2. This will affect how 
%                          eight-second science values are unscaled. 
%
%  Optional arguments:
%
%    write_mat_file        1 indicates the output data structures should be dumped to 
%                          a .mat file, 0 indicates do not dump. Default is 0. 
%    output_directory      The output location. Default is working directory
%    bit_trim_in_filename  Name of an H5 file containing bit-trim parameters, or 
%                          an empty string to use bit-trim values from an
%                          engineering packet in the data stream. Default is 
%                          an empty string.
%    bit_trim_out_filename Name of an H5 file to dump bit trim values to, from 
%                          engineering packets in the data stream, or an empty 
%                          string to not dump bit trim values. Default is an empty
%                          string. Cannot specify both a bit trim in and out file
%    packet_format         0 indicates packet is pre-Aug 2005
%                          and the packet length in the packet header is wrong.
%                          1 indicates post-Aug 2005. Default is 1.
%
%  Return values (data structure formats are described in the README):
%
%    ifgs                  Structure containing unpacked interferograms and 
%                          associated data (FOR, sweep_dir, time)
%    metadata              Structure containing metadata
%    packets               Structure containing fields from packets, broken out and
%                          unscaled as described in the XML database
%
%---------------------------------------------------------------
function [ifgs, metadata, packets] = rdr_convert(packets_filename, xml_database_version, side, varargin)

set_globals

ifgs = [];
metadata = [];
packets = [];
    
RDR_CONVERT_VERSION='1.2.0';  % FIXME: set to the current version number before releasing. For example
                                 %   'x.y.z' or 'x.y.z beta'

XML_DATABASE_VERSIONS={'1.56', '1.69', 'draft_j1', 'j1_55_005', 'j1_55a_008', 'j2_OO_12'};  % supported XML database versions

% add the readers source directory to the path
source_filename = which('rdr_convert.m');
try
    % pre-R2010b form
    [pathstr,name,ext,versn] = fileparts(source_filename); 
catch err
    % R2010b and later
    [pathstr,name,ext] = fileparts(source_filename); 
end
addpath([pathstr, filesep, 'readers']);

% parse arguments
ip = inputParser;
ip.addRequired('packets_filename', @ischar);
ip.addRequired('xml_database_version', @ischar);
ip.addRequired('side', @isnumeric);
ip.addOptional('write_mat_file', 0, @isnumeric);
ip.addOptional('output_dir', pwd, @ischar);
ip.addOptional('bit_trim_in_filename', '', @ischar);
ip.addOptional('bit_trim_out_filename', '', @ischar);
ip.addOptional('packet_format', 1, @isnumeric);
ip.parse(packets_filename, xml_database_version, side, varargin{:});

if ~isempty(ip.Results.bit_trim_in_filename) && ~isempty(ip.Results.bit_trim_out_filename)
    fprintf('\nError: cannot specify both arguments bit_trim_in_filename and bit_trim_out_filename\n')
    return
end

if ~ismember([xml_database_version], XML_DATABASE_VERSIONS)
    fprintf(['\nError: XML database version not supported: %s\nSpecify ' ...
             'one of these XML database versions:\n'], xml_database_version)
    fprintf('  %s\n', XML_DATABASE_VERSIONS{:})
    return
end

if side ~= 1 && side ~= 2
    fprintf('\nError: must specify side 1 or side 2.\n')
    return
end

warning off backtrace % don't print backtrace on warnings

fprintf('\nrdr_convert, version %s\n', RDR_CONVERT_VERSION)

% set metadata
g_metadata.user_options = ip.Results;
g_metadata.rdr_convert_version =  RDR_CONVERT_VERSION;

% construct output filenames
try
    % pre-R2010b form
    [pathstr, name, ext, versn] = fileparts(g_metadata.user_options.packets_filename);
catch err
    % R2010b and later
    [pathstr, name, ext] = fileparts(g_metadata.user_options.packets_filename);
end
matfilename = [g_metadata.user_options.output_dir, filesep, name, '.mat'];

% read packets from input file
fprintf('\nReading input...\n')
fprintf('  Input RDR file: %s\n', g_metadata.user_options.packets_filename)
fprintf('  XML database version: %s\n', xml_database_version)
fprintf('  Side: %d\n', side)
read_tic_id = tic;
read_input(g_metadata.user_options.packets_filename, g_metadata.user_options.packet_format, xml_database_version, side);
read_time = toc(read_tic_id);

% print packet counts to screen
print_packet_counts();

% bit_unpack interferograms
fprintf('\nUnpacking interferograms...\n')
unpack_tic_id = tic;
[err,  g_metadata.bit_trim_info] = unpack_ifgs(g_metadata.user_options.bit_trim_in_filename, g_metadata.user_options.bit_trim_out_filename);
if(err ~= 0)
    fprintf(['  Error unpacking interferograms. Interferogram RDR data will only be saved in packed form.\n'])
end

% set return variables from globals
ifgs = g_ifgs;
metadata = g_metadata;
packets = g_packets;

unpack_time = toc(unpack_tic_id);

% save some variables to matlab file
if(g_metadata.user_options.write_mat_file)
  fprintf('\nWriting Matlab output...\n')
  fprintf('  Output Matlab file: %s\n', matfilename)
  write_mat_tic_id = tic;  
  save_list = {'ifgs', 'metadata', 'packets'};
  save(matfilename, save_list{:}, '-mat')
  write_mat_time = toc(write_mat_tic_id);
end
    
% $$$ % format output and dump to cdf
% $$$ if(dump_netcdf)
% $$$   fprintf('\nWriting NetCDF output...\n')
% $$$   fprintf('  Output NetCDF file: %s\n', outfilename)
% $$$   write_nc_tic_id = tic;
% $$$   try
% $$$       write_output(outfilename);
% $$$   catch err
% $$$       fprintf(['  Failed to write NetCDF output! Note that there is ' ...
% $$$                'a known issue writing NetCDF with Matlab versions ' ...
% $$$                'from 2010 and later\n'], outfilename)
% $$$   end
% $$$   write_nc_time = toc(write_nc_tic_id);
% $$$ end
    
fprintf('\nElapsed time:\n')
fprintf('  reading file:              %f seconds\n', read_time)
fprintf('  unpacking interferograms:  %f seconds\n', unpack_time)
if(g_metadata.user_options.write_mat_file)
  fprintf('  writing Matlab file:       %f seconds\n', write_mat_time)
end
% $$$ if(dump_netcdf)
% $$$   fprintf('  writing NetCDF file:       %f seconds\n', write_nc_time)
% $$$ end
  
clear_globals

end

%-------------------------------------------------------------
%
% function read_input:
%   read packets from files
%
%-------------------------------------------------------------
function read_input(filename, packet_format, xml_database_version, side)
    
set_globals

% initialize packet structures
initialize_packet_structures();

% set up pointers to read functions
build_read_function_pointer_table(xml_database_version, side);

num_packets=0;  
packet_read_error=0;

if(packet_format == 0)
    warning('User specified packet format 0, indicating data is from before August 2005. Will attempt to correct packet lengths in packet headers.\n')
end
    
%tic % set a timer

% open a binary file
[fid, message] = fopen(filename,'r','b');

% if file error, print message and stop
if(fid == -1)
    error('Error opening file %s: %s', filename, message)
end

% read all the packets in the file
% read first header pair
is_not_eof=read_packet_headers(fid, packet_format);

% initialize packet stats
%packet_range=1:50000;
%g_metadata.packet_stats.type(packet_range)=0;
%g_metadata.packet_stats.length(packet_range)=0;
%g_metadata.packet_stats.time(packet_range)=0;
%g_metadata.packet_stats.position(packet_range)=0;

% read remaining packets in the file
while(is_not_eof);
    
    % if packet header is not empty
    if (~isempty(g_header.apid))
        
        % read packet
        if(g_header.apid < 1280 || g_header.apid > 1400)
            packet_read_error=read_unknown_packet(fid);
        else
            g_read_function_pointer_table{g_header.apid - 1279}(fid);
        end
        
        % check for errors
        if(packet_read_error==2)
            warning('Read error: lost sync reading unknown packet, apid = %d', g_header.apid)
            warning('Aborting read')
            return
        end
        
        num_packets=num_packets+1;
        
        g_metadata.packet_stats.apid(num_packets)=g_header.apid;
        g_metadata.packet_stats.time(num_packets)=g_header.time;
        g_metadata.packet_stats.start_byte(num_packets)=g_header.start_byte;
        
    % else packet header is empty
    else
        warning('Packet header is empty')
    end;
        
    % read the next header to see if at eof
    is_not_eof=read_packet_headers(fid, packet_format);
    
end; % end WHILE NOT EOF

% finished reading all packets in the file
fclose(fid);
fprintf('  Number of packets read: %d\n', num_packets)

%time = toc; % How long to read the file?
%fprintf('  Elapsed time reading file: %f seconds\n', time)

end % read_input

%-------------------------------------------------------------
%
% function print_packet_count:
%   print packet stats to screen
%
%-------------------------------------------------------------
function print_packet_counts()

set_globals

fprintf('\n  Packet counts by type:\n')
fprintf('  INTERFEROGRAM RDR')
fprintf('\n      (FOV)   (1)  (2)  (3)  (4)  (5)  (6)  (7)  (8)  (9)')
%fprintf('\n             ---- ---- ---- ---- ---- ---- ---- ---- ----')
for scene = {'es', 'ds', 'ict'}
    for band = {'lw', 'mw', 'sw'}
        fprintf('\n    %s%4s:  ', upper(char(band)), upper(char(scene)))
        for fov = 1:9
            fprintf('%4d ', g_metadata.packet_stats.num_packets.(strcat('ifg_',char(band),'_fov', int2str(fov),'_',char(scene))))
        end
    end
end
fprintf('\n  DIAGNOSTIC RDR')
fprintf('\n    LW DIAG:                  %d', g_metadata.packet_stats.num_packets.lw_diag)
fprintf('\n    MW DIAG:                  %d', g_metadata.packet_stats.num_packets.mw_diag)
fprintf('\n    SW DIAG:                  %d', g_metadata.packet_stats.num_packets.sw_diag)
fprintf('\n  FOUR MINUTE ENGINEERING:    %d', g_metadata.packet_stats.num_packets.four_min_eng)
fprintf('\n  EIGHT SECOND SCIENCE:       %d', g_metadata.packet_stats.num_packets.eight_sec_sci)
fprintf('\n  HK TELEMETRY RDR')
fprintf('\n    subpacket 1:              %d', g_metadata.packet_stats.num_packets.instr_hk1)
fprintf('\n    subpacket 2:              %d', g_metadata.packet_stats.num_packets.instr_hk2)
fprintf('\n    subpacket 3:              %d', g_metadata.packet_stats.num_packets.instr_hk3)
fprintf('\n    subpacket 4:              %d', g_metadata.packet_stats.num_packets.instr_hk4)
fprintf('\n    subpacket 5:              %d', g_metadata.packet_stats.num_packets.instr_hk5)
fprintf('\n    subpacket 6:              %d', g_metadata.packet_stats.num_packets.instr_hk6)
fprintf('\n    subpacket 7:              %d', g_metadata.packet_stats.num_packets.instr_hk7)
fprintf('\n    subpacket 8:              %d', g_metadata.packet_stats.num_packets.instr_hk8)
fprintf('\n  UNKNOWN:                    %d\n', g_metadata.packet_stats.num_packets.unknown)

end
